# -*- coding: utf-8 -*-
# Copyright © 2007 Lateef Alabi-Oki
#
# This file is part of striim.
#
# striim is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# striim is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with striim; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
# USA

"""
This module documents a class that fetches album images from amazon.

@author: Lateef Alabi-Oki
@organization: The striim Project
@copyright: Copyright © 2007 Lateef Alabi-Oki
@license: GNU GPLv2 or Later
@contact: mystilleef@gmail.com
"""

query = "http://webservices.amazon.com/onca/xml?\
Service=AWSECommerceService&AWSAccessKeyId=0SEJ8DTEWT1YHS4E4HG2\
&Operation=ItemSearch\
&SearchIndex=Music\
&Keywords=%s&\
Artist=%s&\
ResponseGroup=Images"

class AlbumFetcher(object):
	"""
	This class fetches album art cover images from amazon.
	"""

	def __init__(self, manager, player):
		"""
		Initialize object.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param manager: Reference to the Manager instance.
		@type manager: A Manager object.

		@param player: Reference to the Player instance.
		@type player: A Player object.
		"""
		self.__init_attributes(manager, player)
		self.__precompile_methods()
		self.__check_album_folder()
		self.__signal_id_1 = manager.connect("destroy", self.__destroy_cb)
		self.__signal_id_2 = player.connect("album-name-found", self.__album_name_found_cb)
		self.__signal_id_3 = player.connect("no-album-name-found", self.__no_album_name_found_cb)

	def __init_attributes(self, manager, player):
		"""
		Initialize data attributes.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param manager: Reference to the Manager instance.
		@type manager: A Manager object.
		"""
		self.__player = player
		self.__result = ""
		self.__title = None
		self.__handle = None
		self.__manager = manager
		self.__album_folder = None
		self.__signal_id_1 = self.__signal_id_2 = self.__signal_id_3 = None
		return

	def __process_metadata(self, artist, title):
		"""
		Process song metadata.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param artist: Artist name.
		@type artist: A String object.

		@param title: A title or album.
		@type title: A String object.
		"""
		from Exceptions import AlbumImageExistError
		try:
			self.__reset_attributes()
			self.__set_image_name(title)
			self.__check_album_image()
			self.__search_amazon(artist, title)
		except AlbumImageExistError:
			pass
		return

	def __check_album_image(self):
		"""
		Check if an album image for a title exists.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.
		"""
		from os import path
		download_location = path.join(self.__album_folder, self.__title)
		from operator import not_
		if not_(path.exists(download_location)): return
		self.__player.emit("album-image-found", download_location)
		from Exceptions import AlbumImageExistError
		raise AlbumImageExistError
		return

	def __get_image_location(self, xml_result):
		"""
		Get location of album image cover art.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param xml_result: The result of the amazon search query.
		@type xml_result: A String object.

		@return: Location of an album art image.
		@rtype: A String object.
		"""
		from xml.parsers.expat import ExpatError
		from Exceptions import AlbumImageError
		try:
			self.__player.response()
			from xml.dom.minidom import parseString
			xml_object = parseString(xml_result)
			root_element = xml_object.documentElement
			nodes = root_element.getElementsByTagName("MediumImage")
			if not nodes: raise AlbumImageError
			image = nodes[0].firstChild.firstChild.nodeValue
			xml_object.unlink()
			self.__player.response()
		except ExpatError:
			raise AlbumImageError
		return image

	def __check_album_folder(self):
		"""
		Create the album images folder if it does not exist.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.
		"""
		from Striim.Global import metadata_folder
		from os import path, makedirs
		album_folder = path.join(metadata_folder, "album")
		self.__album_folder = album_folder
		if path.exists(album_folder): return
		makedirs(album_folder)
		return

	def __set_image_name(self, title):
		"""
		Define the name of the album image.

		@param title: Title or album of a song.
		@type title: A String object.
		"""
		title = title.strip()
		title = title.replace("/", "-")
		title = title.replace(" ", "_")
		self.__title = title + ".jpg"
		return

	def __reset_attributes(self):
		"""
		Reset key varialbe before fetching new album cover.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.
		"""
		self.__result = ""
		if self.__handle: self.__handle.cancel()
		self.__handle = None
		return

	def __error(self):
		"""
		Emit error message.

		@param self: Reference to the Parser instance.
		@type self: A Parser object.
		"""
		self.__player.response()
		self.__player.emit("no-album-image-found")
		if self.__handle: self.__handle.cancel()
		self.__result = ""
		self.__handle = None
		self.__player.response()
		print "Error: No shit!"
		return

	def __search_amazon(self, artist, title):
		"""
		Search amazon for album cover art.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param artist: Artist name.
		@type artist: A String object.

		@param title: Title or Album name.
		@type title: A String object.
		"""
		try:
			self.__player.response()
			from urllib import quote
			uri = query % (quote(title), quote(artist))
			from gnomevfs import OPEN_READ, URI
			from gnomevfs.async import open
			self.__reset_attributes()
			open(URI(uri), self.__open_cb, OPEN_READ, 10)
			self.__player.response()
		except:
			self.__error()
		return

	def __open_cb(self, handle, result):
		"""
		Callback to the async open function.

		@param self: Reference to the Parser instance.
		@type self: A Parser object.
		"""
		try:
			self.__player.response()
			self.__handle = handle
			handle.read(4096, self.__read_cb)
		except:
			self.__error()
		return

	def __read_cb(self, handle, buffer, result, bytes):
		"""
		Callback to the async open function.

		@param self: Reference to the Parser instance.
		@type self: A Parser object.
		"""
		from gnomevfs import EOFError
		from operator import is_, eq, contains, not_
		self.__player.response()
		if is_(result, None):
			try:
				self.__result += buffer
				handle.read(4096, self.__read_cb)
			except:
				self.__error()
		elif eq(result, EOFError):
			try:
				handle.close(self.__close_cb)
			except:
				self.__error()
		else:
			try:
				handle.close(self.__close_cb)
			except:
				self.__error()
		return

	def __close_cb(self, *args):
		"""
		Callback to the async read function.

		@param self: Reference to the Parser instance.
		@type self: A Parser object.
		"""
		from Exceptions import AlbumImageError
		try:
			self.__player.response()
			image_uri = self.__get_image_location(self.__result)
			self.__check_album_folder()
			self.__download_image(image_uri)
		except AlbumImageError:
			self.__player.emit("no-album-image-found")
		return

	def __download_image(self, image_uri):
		"""
		Download album art image for image location

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param image_uri: Location of image.
		@type image_uri: A String object.
		"""
		from os import path
		from gnomevfs import make_uri_from_shell_arg
		from gnomevfs import XFER_OVERWRITE_MODE_REPLACE
		from gnomevfs import XFER_ERROR_MODE_QUERY, URI
		from gnomevfs.async import xfer
		XFER_TARGET_DEFAULT_PERMS = 1 << 12
		try:
			download_location = path.join(self.__album_folder, self.__title)
			download_uri = make_uri_from_shell_arg(download_location)
			self.__player.response()
			xfer(source_uri_list=[URI(image_uri)],
					target_uri_list=[URI(download_uri)],
					xfer_options=XFER_TARGET_DEFAULT_PERMS,
					error_mode=XFER_ERROR_MODE_QUERY,
					priority = 10,
					overwrite_mode=XFER_OVERWRITE_MODE_REPLACE,
					progress_update_callback=self.__update_cb,
					update_callback_data=download_location,
					progress_sync_callback=self.__sync_cb)
		except:
			self.__error()
		return

	def __sync_cb(self, info):
		"""
		Callback the GNOME-VFS asynchronous transfer method.

		@param self: Reference to the FileSaver instance.
		@type self: A FileSaver object.
		"""
		self.__player.response()
		if info.vfs_status:
			return False
		return  True

	def __update_cb(self, handle, info, data):
		"""
		Callback to the GNOME-VFS asynchronous transfer method.

		@param self: Reference to the FileSaver instance.
		@type self: A FileSaver object.
		"""
		try:
			self.__player.response()
			self.__handle = handle
			if info.vfs_status: raise Exception
			from gnomevfs import XFER_PHASE_COMPLETED
			from operator import ne, not_
			if ne(info.phase, XFER_PHASE_COMPLETED): return True
			from os import path
			if not_(path.exists(data)):
				self.__player.emit("no-album-image-found")
			else:
				self.__player.emit("album-image-found", data)
		except:
			self.__error()
		return True

	def __album_name_found_cb(self, player, data):
		"""
		Handles callback when the "album-name-found" signal is emitted.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param player: Reference to the Player instance.
		@type player: A Player object.

		@param data: Metadata information about a track.
		@type data: A Tuple object.
		"""
		self.__player.response()
		self.__reset_attributes()
		artist, title, album = data
		self.__process_metadata(artist, album)
		self.__player.response()
		return

	def __no_album_name_found_cb(self, player, data):
		"""
		Handles callback when the "album-name-found" signal is emitted.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.

		@param player: Reference to the Player instance.
		@type player: A Player object.

		@param data: Metadata information about a track.
		@type data: A Tuple object.
		"""
		try:
			self.__player.response()
			self.__reset_attributes()
			artist, title = data
			self.__process_metadata(artist, title)
			self.__player.response()
		except ValueError:
			pass
		return

	def __destroy_cb(self, *args):
		"""
		Handles callback when the "destroy" signal is emitted.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.
		"""
		from Striim.utils import disconnect_signal, delete_attributes
		disconnect_signal(self.__signal_id_1, self.__manager)
		disconnect_signal(self.__signal_id_2, self.__player)
		disconnect_signal(self.__signal_id_3, self.__player)
		delete_attributes(self)
		del self
		self = None
		return

	def __precompile_methods(self):
		"""
		Optimize psyco methods.

		@param self: Reference to the AlbumFetcher instance.
		@type self: An AlbumFetcher object.
		"""
		try:
			from psyco import bind
			bind(self.__download_image)
			bind(self.__search_amazon)
			bind(self.__process_metadata)
		except ImportError:
			pass
		return
